@@ -101,7 +101,9 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
 
 (defun exwm-input--update-focus ()
   "Update input focus."
-  (when (window-live-p exwm-input--focus-window)
+  (when (and (window-live-p exwm-input--focus-window)
+             ;; Do not update input focus when there's an active minibuffer.
+             (not (active-minibuffer-window)))
     (with-current-buffer (window-buffer exwm-input--focus-window)
       (if (eq major-mode 'exwm-mode)
           (if (not (eq exwm--frame exwm-workspace--current))
@@ -308,6 +310,10 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
                  (setq event (xcb:keysyms:keysym->event exwm--connection
                                                         keysym state)))
         (when (eq major-mode 'exwm-mode)
+          ;; FIXME: This functionality seems not working, e.g. when this
+          ;;        command would activate the minibuffer, the temporary
+          ;;        line-mode would actually quit before the minibuffer
+          ;;        becomes active.
           (setq exwm-input--temp-line-mode t
                 exwm-input--during-key-sequence t)
           (exwm-input--grab-keyboard))  ;grab keyboard temporarily
@@ -318,6 +324,34 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
                      :time xcb:Time:CurrentTime))
   (xcb:flush exwm--connection))
 
+(defun exwm-input--update-mode-line (id)
+  "Update the propertized `mode-line-process' for window ID."
+  (let (help-echo cmd mode)
+    (cl-case exwm--on-KeyPress
+      ((exwm-input--on-KeyPress-line-mode)
+       (setq mode "line"
+             help-echo "mouse-1: Switch to char-mode"
+             cmd `(lambda ()
+                    (interactive)
+                    (exwm-input-release-keyboard ,id))))
+      ((exwm-input--on-KeyPress-char-mode)
+       (setq mode "char"
+             help-echo "mouse-1: Switch to line-mode"
+             cmd `(lambda ()
+                    (interactive)
+                    (exwm-input-grab-keyboard ,id)))))
+    (with-current-buffer (exwm--id->buffer id)
+      (setq mode-line-process
+            `(": "
+              (:propertize ,mode
+                           help-echo ,help-echo
+                           mouse-face mode-line-highlight
+                           local-map
+                           (keymap
+                            (mode-line
+                             keymap
+                             (down-mouse-1 . ,cmd)))))))))
+
 (defun exwm-input--grab-keyboard (&optional id)
   "Grab all key events on window ID."
   (unless id (setq id (exwm--buffer->id (window-buffer))))
@@ -325,15 +359,14 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
     (when (xcb:+request-checked+request-check exwm--connection
               (make-instance 'xcb:GrabKey
                              :owner-events 0
-                             :grab-window
-                             (with-current-buffer (exwm--id->buffer id)
-                               exwm--container)
+                             :grab-window id
                              :modifiers xcb:ModMask:Any
                              :key xcb:Grab:Any
                              :pointer-mode xcb:GrabMode:Async
                              :keyboard-mode xcb:GrabMode:Sync))
       (exwm--log "Failed to grab keyboard for #x%x" id))
-    (setq exwm--on-KeyPress #'exwm-input--on-KeyPress-line-mode)))
+    (with-current-buffer (exwm--id->buffer id)
+      (setq exwm--on-KeyPress #'exwm-input--on-KeyPress-line-mode))))
 
 (defun exwm-input--release-keyboard (&optional id)
   "Ungrab all key events on window ID."
@@ -342,46 +375,31 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
     (when (xcb:+request-checked+request-check exwm--connection
               (make-instance 'xcb:UngrabKey
                              :key xcb:Grab:Any
-                             :grab-window
-                             (with-current-buffer (exwm--id->buffer id)
-                               exwm--container)
+                             :grab-window id
                              :modifiers xcb:ModMask:Any))
       (exwm--log "Failed to release keyboard for #x%x" id))
-    (setq exwm--on-KeyPress #'exwm-input--on-KeyPress-char-mode)))
+    (with-current-buffer (exwm--id->buffer id)
+      (setq exwm--on-KeyPress #'exwm-input--on-KeyPress-char-mode))))
 
 ;;;###autoload
 (defun exwm-input-grab-keyboard (&optional id)
   "Switch to line-mode."
-  (interactive)
-  (exwm-input--grab-keyboard id)
-  (setq mode-line-process
-        '(": "
-          (:propertize "line"
-                       help-echo "mouse-1: Switch to char-mode"
-                       mouse-face mode-line-highlight
-                       local-map
-                       (keymap
-                        (mode-line
-                         keymap
-                         (down-mouse-1 . exwm-input-release-keyboard))))))
-  (force-mode-line-update))
+  (interactive (list (exwm--buffer->id (window-buffer))))
+  (when id
+    (with-current-buffer (exwm--id->buffer id)
+      (exwm-input--grab-keyboard id)
+      (exwm-input--update-mode-line id)
+      (force-mode-line-update))))
 
 ;;;###autoload
 (defun exwm-input-release-keyboard (&optional id)
   "Switch to char-mode."
-  (interactive)
-  (exwm-input--release-keyboard id)
-  (setq mode-line-process
-        '(": "
-          (:propertize "char"
-                       help-echo "mouse-1: Switch to line-mode"
-                       mouse-face mode-line-highlight
-                       local-map
-                       (keymap
-                        (mode-line
-                         keymap
-                         (down-mouse-1 . exwm-input-grab-keyboard))))))
-  (force-mode-line-update))
+  (interactive (list (exwm--buffer->id (window-buffer))))
+  (when id
+    (with-current-buffer (exwm--id->buffer id)
+      (exwm-input--release-keyboard id)
+      (exwm-input--update-mode-line id)
+      (force-mode-line-update))))
 
 (defun exwm-input--fake-key (event)
   "Fake a key event equivalent to Emacs event EVENT."
@@ -464,6 +482,14 @@ SIMULATION-KEYS is an alist of the form (original-key . simulated-key)."
         (dolist (j pair)
           (exwm-input--fake-key j))))))
 
+(defun exwm-input--on-pre-command ()
+  "Run in `pre-command-hook'."
+  (setq exwm-input--during-command t))
+
+(defun exwm-input--on-post-command ()
+  "Run in `post-command-hook'."
+  (setq exwm-input--during-command nil))
+
 (declare-function exwm-floating--stop-moveresize "exwm-floating.el"
                   (&rest _args))
 (declare-function exwm-floating--do-moveresize "exwm-floating.el"
@@ -498,14 +524,20 @@ SIMULATION-KEYS is an alist of the form (original-key . simulated-key)."
   ;; `pre-command-hook' marks the end of a key sequence (existing or not)
   (add-hook 'pre-command-hook #'exwm-input--finish-key-sequence)
   ;; Control `exwm-input--during-command'
-  (add-hook 'pre-command-hook (lambda () (setq exwm-input--during-command t)))
-  (add-hook 'post-command-hook
-            (lambda () (setq exwm-input--during-command nil)))
+  (add-hook 'pre-command-hook #'exwm-input--on-pre-command)
+  (add-hook 'post-command-hook #'exwm-input--on-post-command)
   ;; Update focus when buffer list updates
   (add-hook 'buffer-list-update-hook #'exwm-input--on-buffer-list-update)
   ;; Update prefix keys for global keys
   (exwm-input--update-global-prefix-keys))
 
+(defun exwm-input--exit ()
+  "Exit the input module."
+  (remove-hook 'pre-command-hook #'exwm-input--finish-key-sequence)
+  (remove-hook 'pre-command-hook #'exwm-input--on-pre-command)
+  (remove-hook 'post-command-hook #'exwm-input--on-post-command)
+  (remove-hook 'buffer-list-update-hook #'exwm-input--on-buffer-list-update))
+
 
 
 (provide 'exwm-input)
