@@ -103,7 +103,8 @@
       (xcb:+request exwm--connection
           (make-instance 'xcb:icccm:set-WM_STATE
                          :window id :state xcb:icccm:WM_STATE:NormalState
-                         :icon xcb:Window:None)))
+                         :icon xcb:Window:None))
+      (setq exwm-state xcb:icccm:WM_STATE:NormalState))
     (xcb:+request exwm--connection
         (make-instance 'xcb:SendEvent
                        :propagate 0 :destination id
@@ -122,27 +123,27 @@
 
 (defun exwm-layout--hide (id)
   "Hide window ID."
-  (unless (eq xcb:icccm:WM_STATE:IconicState ;already hidden
-              (with-current-buffer (exwm--id->buffer id) exwm-state))
-    (exwm--log "Hide #x%x" id)
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:ChangeWindowAttributes
-                       :window id :value-mask xcb:CW:EventMask
-                       :event-mask xcb:EventMask:NoEvent))
-    (xcb:+request exwm--connection (make-instance 'xcb:UnmapWindow :window id))
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:ChangeWindowAttributes
-                       :window id :value-mask xcb:CW:EventMask
-                       :event-mask exwm--client-event-mask))
-    (with-current-buffer (exwm--id->buffer id)
+  (with-current-buffer (exwm--id->buffer id)
+    (unless (eq xcb:icccm:WM_STATE:IconicState exwm-state) ;already hidden
+      (exwm--log "Hide #x%x" id)
       (xcb:+request exwm--connection
-          (make-instance 'xcb:UnmapWindow :window exwm--container)))
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:icccm:set-WM_STATE
-                       :window id
-                       :state xcb:icccm:WM_STATE:IconicState
-                       :icon xcb:Window:None))
-    (xcb:flush exwm--connection)))
+          (make-instance 'xcb:ChangeWindowAttributes
+                         :window id :value-mask xcb:CW:EventMask
+                         :event-mask xcb:EventMask:NoEvent))
+      (xcb:+request exwm--connection (make-instance 'xcb:UnmapWindow :window id))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:ChangeWindowAttributes
+                         :window id :value-mask xcb:CW:EventMask
+                         :event-mask exwm--client-event-mask))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:UnmapWindow :window exwm--container))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:icccm:set-WM_STATE
+                         :window id
+                         :state xcb:icccm:WM_STATE:IconicState
+                         :icon xcb:Window:None))
+      (setq exwm-state xcb:icccm:WM_STATE:IconicState)
+      (xcb:flush exwm--connection))))
 
 (defvar exwm-workspace--current)
 (defvar exwm-workspace--list)
@@ -164,13 +165,19 @@
     (exwm-layout--resize-container exwm--id exwm--container 0 0
                                    (exwm-workspace--current-width)
                                    (exwm-workspace--current-height))
+    ;; Raise the X window.
+    (xcb:+request exwm--connection
+        (make-instance 'xcb:ConfigureWindow
+                       :window exwm--container
+                       :value-mask xcb:ConfigWindow:StackMode
+                       :stack-mode xcb:StackMode:Above))
     (xcb:+request exwm--connection
         (make-instance 'xcb:ewmh:set-_NET_WM_STATE
                        :window exwm--id
                        :data (vector xcb:Atom:_NET_WM_STATE_FULLSCREEN)))
     (xcb:flush exwm--connection)
     (setq exwm--fullscreen t)
-    (exwm-input-release-keyboard)))
+    (call-interactively #'exwm-input-release-keyboard)))
 
 ;;;###autoload
 (defun exwm-layout-unset-fullscreen (&optional id)
@@ -179,22 +186,31 @@
   (with-current-buffer (if id (exwm--id->buffer id) (window-buffer))
     (unless exwm--fullscreen
       (user-error "Not in full-screen mode."))
-    ;; Restore the floating frame if the client is floating
-    (when exwm--floating-frame
+    (if exwm--floating-frame
+        ;; Restore the floating frame.
+        (xcb:+request exwm--connection
+            (make-instance 'xcb:ConfigureWindow
+                           :window exwm--container
+                           :value-mask (eval-when-compile
+                                         (logior xcb:ConfigWindow:X
+                                                 xcb:ConfigWindow:Y))
+                           :x (elt exwm--floating-frame-position 0)
+                           :y (elt exwm--floating-frame-position 1)))
+      ;; Put the X window just above the Emacs frame.
       (xcb:+request exwm--connection
           (make-instance 'xcb:ConfigureWindow
                          :window exwm--container
-                         :value-mask (eval-when-compile
-                                       (logior xcb:ConfigWindow:X
-                                               xcb:ConfigWindow:Y))
-                         :x (elt exwm--floating-frame-position 0)
-                         :y (elt exwm--floating-frame-position 1))))
+                         :value-mask (logior xcb:ConfigWindow:Sibling
+                                             xcb:ConfigWindow:StackMode)
+                         :sibling (frame-parameter exwm-workspace--current
+                                                   'exwm-container)
+                         :stack-mode xcb:StackMode:Above)))
     (exwm-layout--show exwm--id)
     (xcb:+request exwm--connection
         (make-instance 'xcb:ewmh:set-_NET_WM_STATE :window exwm--id :data []))
     (xcb:flush exwm--connection)
     (setq exwm--fullscreen nil)
-    (exwm-input-grab-keyboard)))
+    (call-interactively #'exwm-input-grab-keyboard)))
 
 (defvar exwm-layout--fullscreen-frame-count 0
   "Count the fullscreen workspace frames.")
@@ -224,33 +240,60 @@
       (xcb:flush exwm--connection)))
   (cl-incf exwm-layout--fullscreen-frame-count))
 
+(defvar exwm-layout--other-buffer-exclude-exwm-mode-buffers nil
+  "When non-nil, prevent EXWM buffers from being selected by `other-buffer'.")
+
+(defvar exwm-layout--other-buffer-exclude-buffers nil
+  "List of buffers that should not be selected by `other-buffer'.")
+
+(defun exwm-layout--other-buffer-predicate (buffer)
+  "Return non-nil when the BUFFER may be displayed in selected frame.
+
+Prevents EXWM-mode buffers already being displayed on some other window from
+being selected.
+
+Should be set as `buffer-predicate' frame parameter for all
+frames.  Used by `other-buffer'.
+
+When variable `exwm-layout--other-buffer-exclude-exwm-mode-buffers'
+is t EXWM buffers are never selected by `other-buffer'.
+
+When variable `exwm-layout--other-buffer-exclude-buffers' is a
+list of buffers, EXWM buffers belonging to that list are never
+selected by `other-buffer'."
+  (or (not (eq 'exwm-mode (buffer-local-value 'major-mode buffer)))
+      (and (not exwm-layout--other-buffer-exclude-exwm-mode-buffers)
+           (not (memq buffer exwm-layout--other-buffer-exclude-buffers))
+           ;; Do not select if already shown in some window.
+           (not (get-buffer-window buffer t)))))
+
 (defvar exwm-layout-show-all-buffers nil
   "Non-nil to allow switching to buffers on other workspaces.")
 
 (defun exwm-layout--refresh ()
   "Refresh layout."
   (let ((frame (selected-frame))
-        (placeholder (or (get-buffer "*scratch*")
-                         (progn
-                           (set-buffer-major-mode
-                            (get-buffer-create "*scratch*"))
-                           (get-buffer "*scratch*"))))
+        covered-buffers             ;EXWM-buffers covered by a new X window.
+        vacated-windows             ;Windows previously displaying EXWM-buffers.
         windows)
     (if (not (memq frame exwm-workspace--list))
         (if (frame-parameter frame 'exwm-outer-id)
             ;; Refresh a floating frame
-            (when (eq major-mode 'exwm-mode)
-              (let ((window (frame-first-window frame)))
-                (with-current-buffer (window-buffer window)
+            (let ((window (frame-first-window frame)))
+              (with-current-buffer (window-buffer window)
+                (when (and (eq major-mode 'exwm-mode)
+                           ;; It may be a buffer waiting to be killed.
+                           (exwm--id->buffer exwm--id))
                   (exwm--log "Refresh floating window #x%x" exwm--id)
                   (exwm-layout--show exwm--id window))))
           ;; Other frames (e.g. terminal/graphical frame of emacsclient)
           ;; We shall bury all `exwm-mode' buffers in this case
           (setq windows (window-list frame 0)) ;exclude minibuffer
-          (dolist (window windows)
-            (with-current-buffer (window-buffer window)
-              (when (eq major-mode 'exwm-mode)
-                (set-window-buffer window placeholder)))))
+          (let ((exwm-layout--other-buffer-exclude-exwm-mode-buffers t))
+            (dolist (window windows)
+              (with-current-buffer (window-buffer window)
+                (when (eq major-mode 'exwm-mode)
+                  (switch-to-prev-buffer window))))))
       ;; Refresh the whole workspace
       ;; Workspaces other than the active one can also be refreshed (RandR)
       (exwm--log "Refresh workspace %s" frame)
@@ -264,21 +307,36 @@
             (if (not windows)
                 (when (eq frame exwm--frame) ;for exwm-layout-show-all-buffers
                   (exwm-layout--hide exwm--id))
-              (if (eq frame exwm--frame)
-                  (exwm-layout--show exwm--id (car windows))
-                (exwm-workspace-move-window
-                 (cl-position frame exwm-workspace--list) exwm--id))
               (let ((window (car windows)))
-                ;; Make sure this buffer is not displayed elsewhere
-                (dolist (i (get-buffer-window-list (current-buffer) 0 t))
-                  (unless (eq i window)
-                    (set-window-buffer i placeholder))))))))
+                (if (eq frame exwm--frame)
+                    (exwm-layout--show exwm--id window)
+                  (exwm-workspace-move-window
+                   (cl-position frame exwm-workspace--list) exwm--id))
+                ;; Make sure this buffer is not displayed elsewhere.  Note down
+                ;; windows displaying an EXWM-buffer now displayed elsewhere; we
+                ;; need to display with some other buffer there.
+                (setq vacated-windows
+                      (append vacated-windows (cdr (get-buffer-window-list
+                                                    (current-buffer) 0 t))))
+                ;; Note down when an EXWM-buffer is being covered by this
+                ;; buffer; we don't want it to reappear in some vacated window.
+                (let ((prev-buffer (car-safe
+                                    (car-safe (window-prev-buffers window)))))
+                  (and
+                   prev-buffer
+                   (eq 'exwm-mode (buffer-local-value 'major-mode prev-buffer))
+                   (push prev-buffer covered-buffers))))))))
+      ;; Set some sensible buffer to vacated windows.
+      (let ((exwm-layout--other-buffer-exclude-buffers covered-buffers))
+        (dolist (window vacated-windows)
+          (switch-to-prev-buffer window)))
       ;; Make sure windows floating / on other workspaces are excluded
-      (dolist (window (window-list frame 0))
-        (with-current-buffer (window-buffer window)
-          (when (and (eq major-mode 'exwm-mode)
-                     (or exwm--floating-frame (not (eq frame exwm--frame))))
-            (set-window-buffer window placeholder))))
+      (let ((exwm-layout--other-buffer-exclude-exwm-mode-buffers t))
+        (dolist (window (window-list frame 0))
+          (with-current-buffer (window-buffer window)
+            (when (and (eq major-mode 'exwm-mode)
+                       (or exwm--floating-frame (not (eq frame exwm--frame))))
+              (switch-to-prev-buffer window)))))
       ;; Update _NET_CLIENT_LIST_STACKING
       (xcb:+request exwm--connection
           (make-instance 'xcb:ewmh:set-_NET_CLIENT_LIST_STACKING
@@ -295,16 +353,19 @@
 
 (defun exwm-layout--on-minibuffer-setup ()
   "Refresh layout when minibuffer grows."
-  (run-with-idle-timer 0.01 nil         ;FIXME
-                       (lambda ()
-                         (when (< 1 (window-height (minibuffer-window)))
-                           (exwm-layout--refresh))))
-  ;; Set input focus on the Emacs frame
-  (x-focus-frame (window-frame (minibuffer-selected-window))))
+  (when (frame-parameter nil 'exwm-outer-id)
+    (run-with-idle-timer 0.01 nil         ;FIXME
+                         (lambda ()
+                           (when (< 1 (window-height (minibuffer-window)))
+                             (exwm-layout--refresh))))
+    ;; Set input focus on the Emacs frame
+    (x-focus-frame (window-frame (minibuffer-selected-window)))))
 
 (defun exwm-layout--on-echo-area-change (&optional dirty)
   "Run when message arrives or in `echo-area-clear-hook' to refresh layout."
   (when (and (current-message)
+             ;; Exclude non-graphical frames.
+             (frame-parameter nil 'exwm-outer-id)
              (or (cl-position ?\n (current-message))
                  (> (length (current-message))
                     (frame-width exwm-workspace--current))))
@@ -455,7 +516,8 @@ See also `exwm-layout-enlarge-window'."
                            (window-mode-line-height (frame-root-window
                                                      exwm--floating-frame)))
                         nil t)
-      (exwm-input-grab-keyboard))))
+      (call-interactively #'exwm-input-grab-keyboard))
+    (force-mode-line-update)))
 
 ;;;###autoload
 (defun exwm-layout-toggle-mode-line ()
@@ -466,6 +528,8 @@ See also `exwm-layout-enlarge-window'."
         (exwm-layout-hide-mode-line)
       (exwm-layout-show-mode-line))))
 
+(defvar exwm-layout--timer nil "Timer used to track echo area changes.")
+
 (defun exwm-layout--init ()
   "Initialize layout module."
   ;; Auto refresh layout
@@ -473,9 +537,19 @@ See also `exwm-layout-enlarge-window'."
   (unless (exwm-workspace--minibuffer-own-frame-p)
     ;; Refresh when minibuffer grows
     (add-hook 'minibuffer-setup-hook #'exwm-layout--on-minibuffer-setup t)
-    (run-with-idle-timer 0 t #'exwm-layout--on-echo-area-change t)
+    (setq exwm-layout--timer
+          (run-with-idle-timer 0 t #'exwm-layout--on-echo-area-change t))
     (add-hook 'echo-area-clear-hook #'exwm-layout--on-echo-area-change)))
 
+(defun exwm-layout--exit ()
+  "Exit the layout module."
+  (remove-hook 'window-configuration-change-hook #'exwm-layout--refresh)
+  (remove-hook 'minibuffer-setup-hook #'exwm-layout--on-minibuffer-setup)
+  (when exwm-layout--timer
+    (cancel-timer exwm-layout--timer)
+    (setq exwm-layout--timer nil))
+  (remove-hook 'echo-area-clear-hook #'exwm-layout--on-echo-area-change))
+
 
 
 (provide 'exwm-layout)
