@@ -48,7 +48,7 @@ corresponding buffer.")
 (defun exwm-manage--update-mwm-hints (id &optional force)
   "Update _MOTIF_WM_HINTS."
   (with-current-buffer (exwm--id->buffer id)
-    (unless (and exwm--mwm-hints (not force))
+    (unless (and (not exwm--mwm-hints-decorations) (not force))
       (let ((reply (xcb:+request-unchecked+reply exwm--connection
                        (make-instance 'xcb:icccm:-GetProperty
                                       :window id
@@ -56,7 +56,17 @@ corresponding buffer.")
                                       :type exwm-manage--_MOTIF_WM_HINTS
                                       :long-length 5))))
         (when reply
-          (setq exwm--mwm-hints (append (slot-value reply 'value) nil)))))))
+          ;; Check MotifWmHints.decorations.
+          (with-slots (value) reply
+            (setq value (append value nil))
+            (when (and value
+                       ;; See <Xm/MwmUtil.h> for fields definitions.
+                       (/= 0 (logand
+                              (elt value 0) ;MotifWmHints.flags
+                              2))           ;MWM_HINTS_DECORATIONS
+                       (= 0
+                          (elt value 2))) ;MotifWmHints.decorations
+              (setq exwm--mwm-hints-decorations nil))))))))
 
 (defvar exwm-workspace--current)
 (defvar exwm-workspace--switch-history-outdated)
@@ -90,6 +100,7 @@ corresponding buffer.")
       (exwm--update-class id)
       (exwm--update-transient-for id)
       (exwm--update-normal-hints id)
+      (exwm--update-hints id)
       (exwm-manage--update-geometry id)
       (exwm-manage--update-mwm-hints id)
       ;; No need to manage (please check OverrideRedirect outside)
@@ -99,15 +110,9 @@ corresponding buffer.")
                   (memq xcb:Atom:_NET_WM_WINDOW_TYPE_UTILITY exwm-window-type)
                   (memq xcb:Atom:_NET_WM_WINDOW_TYPE_DIALOG exwm-window-type)
                   (memq xcb:Atom:_NET_WM_WINDOW_TYPE_NORMAL exwm-window-type)))
-             ;; Check _MOTIF_WM_HINTS for Java applications
-             ;; See <Xm/MwmUtil.h> for the definitions of these fields
-             (and exwm--mwm-hints
-                  exwm-instance-name
-                  (/= 0 (logand (elt exwm--mwm-hints 0) ;MotifWmHints.flags
-                                2))             ;MWM_HINTS_DECORATIONS
-                  (= 0 (elt exwm--mwm-hints 2)) ;MotifWmHints.decorations
-                  ;; Java applications only
-                  (string-prefix-p "sun-awt-X11-" exwm-instance-name)
+             ;; Check the _MOTIF_WM_HINTS property.
+             (and (not exwm--mwm-hints-decorations)
+                  (not exwm--hints-input)
                   ;; Floating windows only
                   (or exwm-transient-for exwm--fixed-size
                       (memq xcb:Atom:_NET_WM_WINDOW_TYPE_UTILITY
@@ -129,12 +134,20 @@ corresponding buffer.")
                             exwm-window-type)
                       (memq xcb:Atom:_NET_WM_WINDOW_TYPE_DOCK
                             exwm-window-type))
-            (xcb:+request exwm--connection
-                (make-instance 'xcb:ReparentWindow
-                               :window id
-                               :parent (frame-parameter exwm-workspace--current
-                                                        'exwm-workspace)
-                               :x x :y y)))
+            (let ((frame-geometry (frame-parameter exwm-workspace--current
+                                                   'exwm-geometry))
+                  (workspace (frame-parameter exwm-workspace--current
+                                              'exwm-workspace)))
+              (when (and frame-geometry
+                         (/= x 0)
+                         (/= y 0))
+                (setq x (- x (slot-value frame-geometry 'x))
+                      y (- y (slot-value frame-geometry 'y))))
+              (xcb:+request exwm--connection
+                  (make-instance 'xcb:ReparentWindow
+                                 :window id
+                                 :parent workspace
+                                 :x x :y y))))
           ;; Center window of type _NET_WM_WINDOW_TYPE_SPLASH
           (when (memq xcb:Atom:_NET_WM_WINDOW_TYPE_SPLASH exwm-window-type)
             (xcb:+request exwm--connection
@@ -197,7 +210,6 @@ corresponding buffer.")
                          :data (vconcat (mapcar #'car exwm--id-buffer-alist))))
       (xcb:flush exwm--connection)
       (exwm--update-title id)
-      (exwm--update-hints id)
       (exwm--update-protocols id)
       (exwm--update-state id)
       (if (or exwm-transient-for exwm--fixed-size
@@ -334,56 +346,57 @@ corresponding buffer.")
       (xcb:flush exwm--connection)
       ;; Wait for DestroyNotify event.
       (throw 'return nil))
-    ;; Try to close the X window with WM_DELETE_WINDOW client message.
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:icccm:SendEvent
-                       :destination exwm--id
-                       :event (xcb:marshal
-                               (make-instance 'xcb:icccm:WM_DELETE_WINDOW
-                                              :window exwm--id)
-                               exwm--connection)))
-    (xcb:flush exwm--connection)
-    ;;
-    (unless (memq xcb:Atom:_NET_WM_PING exwm--protocols)
-      ;; The window does not support _NET_WM_PING.  To make sure it'll die,
-      ;; kill it after the time runs out.
-      ;; Hide the container to prevent flickering.
+    (let ((id exwm--id))
+      ;; Try to close the X window with WM_DELETE_WINDOW client message.
       (xcb:+request exwm--connection
-          (make-instance 'xcb:UnmapWindow :window exwm--container))
+          (make-instance 'xcb:icccm:SendEvent
+                         :destination id
+                         :event (xcb:marshal
+                                 (make-instance 'xcb:icccm:WM_DELETE_WINDOW
+                                                :window id)
+                                 exwm--connection)))
       (xcb:flush exwm--connection)
-      (run-with-timer exwm-manage-ping-timeout nil
-                      `(lambda () (exwm-manage--kill-client ,exwm--id)))
-      ;; Wait for DestroyNotify event.
-      (throw 'return nil))
-    ;; Try to determine if the X window is dead with _NET_WM_PING.
-    (setq exwm-manage--ping-lock t)
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:SendEvent
-                       :propagate 0
-                       :destination exwm--id
-                       :event-mask xcb:EventMask:NoEvent
-                       :event (xcb:marshal
-                               (make-instance 'xcb:ewmh:_NET_WM_PING
-                                              :window exwm--id
-                                              :timestamp 0
-                                              :client-window exwm--id)
-                               exwm--connection)))
-    (xcb:flush exwm--connection)
-    (with-timeout (exwm-manage-ping-timeout
-                   (if (yes-or-no-p (format "'%s' is not responding. \
+      ;;
+      (unless (memq xcb:Atom:_NET_WM_PING exwm--protocols)
+        ;; The window does not support _NET_WM_PING.  To make sure it'll die,
+        ;; kill it after the time runs out.
+        ;; Hide the container to prevent flickering.
+        (xcb:+request exwm--connection
+            (make-instance 'xcb:UnmapWindow :window exwm--container))
+        (xcb:flush exwm--connection)
+        (run-with-timer exwm-manage-ping-timeout nil
+                        `(lambda () (exwm-manage--kill-client ,id)))
+        ;; Wait for DestroyNotify event.
+        (throw 'return nil))
+      ;; Try to determine if the X window is dead with _NET_WM_PING.
+      (setq exwm-manage--ping-lock t)
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:SendEvent
+                         :propagate 0
+                         :destination id
+                         :event-mask xcb:EventMask:NoEvent
+                         :event (xcb:marshal
+                                 (make-instance 'xcb:ewmh:_NET_WM_PING
+                                                :window id
+                                                :timestamp 0
+                                                :client-window id)
+                                 exwm--connection)))
+      (xcb:flush exwm--connection)
+      (with-timeout (exwm-manage-ping-timeout
+                     (if (yes-or-no-p (format "'%s' is not responding. \
 Would you like to kill it? "
-                                            (buffer-name)))
-                       (progn (exwm-manage--kill-client exwm--id)
-                              ;; Kill the unresponsive X window and
-                              ;; wait for DestroyNotify event.
-                              (throw 'return nil))
-                     ;; Give up.
-                     (throw 'return nil)))
-      (while (and exwm-manage--ping-lock
-                  (exwm--id->buffer exwm--id)) ;may have been destroyed.
-        (accept-process-output nil 0.1))
-      ;; Give up.
-      (throw 'return nil))))
+                                              (buffer-name)))
+                         (progn (exwm-manage--kill-client id)
+                                ;; Kill the unresponsive X window and
+                                ;; wait for DestroyNotify event.
+                                (throw 'return nil))
+                       ;; Give up.
+                       (throw 'return nil)))
+        (while (and exwm-manage--ping-lock
+                    (exwm--id->buffer id)) ;may have been destroyed.
+          (accept-process-output nil 0.1))
+        ;; Give up.
+        (throw 'return nil)))))
 
 (defun exwm-manage--kill-client (&optional id)
   "Kill an X client."
@@ -409,10 +422,14 @@ Would you like to kill it? "
                          (xcb:+request exwm--connection ,request))))
     (xcb:flush exwm--connection)))
 
+;; FIXME: Make the following values as small as possible.
+(defconst exwm-manage--width-delta-min 5)
+(defconst exwm-manage--height-delta-min 5)
+
 (defun exwm-manage--on-ConfigureRequest (data _synthetic)
   "Handle ConfigureRequest event."
   (let ((obj (make-instance 'xcb:ConfigureRequest))
-        buffer edges)
+        buffer edges width-delta height-delta)
     (xcb:unmarshal obj data)
     (with-slots (window x y width height
                         border-width sibling stack-mode value-mask)
@@ -421,7 +438,31 @@ Would you like to kill it? "
 border-width: %d; sibling: #x%x; stack-mode: %d"
                  window value-mask width height x y
                  border-width sibling stack-mode)
-      (if (setq buffer (exwm--id->buffer window))
+      (if (and (setq buffer (exwm--id->buffer window))
+               (with-current-buffer buffer
+                 (or exwm--fullscreen
+                     ;; Make sure it's a floating X window wanting to resize
+                     ;; itself.
+                     (or (not exwm--floating-frame)
+                         (progn
+                           (setq edges
+                                 (window-inside-pixel-edges
+                                  (get-buffer-window buffer t))
+                                 width-delta (- width (- (elt edges 2)
+                                                         (elt edges 0)))
+                                 height-delta (- height (- (elt edges 3)
+                                                           (elt edges 1))))
+                           ;; We cannot do resizing precisely for now.
+                           (and (if (= 0 (logand value-mask
+                                                 xcb:ConfigWindow:Width))
+                                    t
+                                  (< (abs width-delta)
+                                     exwm-manage--width-delta-min))
+                                (if (= 0 (logand value-mask
+                                                  xcb:ConfigWindow:Height))
+                                    t
+                                  (< (abs height-delta)
+                                     exwm-manage--height-delta-min))))))))
           ;; Send client message for managed windows
           (with-current-buffer buffer
             (setq edges
@@ -446,16 +487,31 @@ border-width: %d; sibling: #x%x; stack-mode: %d"
                                         :height (- (elt edges 3) (elt edges 1))
                                         :border-width 0 :override-redirect 0)
                                        exwm--connection))))
-        (exwm--log "ConfigureWindow (preserve geometry)")
-        ;; Configure the unmanaged window.
-        (xcb:+request exwm--connection
-            (make-instance 'xcb:ConfigureWindow
-                           :window window
-                           :value-mask value-mask
-                           :x x :y y :width width :height height
-                           :border-width border-width
-                           :sibling sibling
-                           :stack-mode stack-mode)))))
+        (if buffer
+            (with-current-buffer buffer
+              (exwm--log "ConfigureWindow (resize floating X window)")
+              (when (and (/= 0 (logand value-mask xcb:ConfigWindow:Width))
+                         (>= (abs width-delta) exwm-manage--width-delta-min))
+                (set-frame-width exwm--floating-frame
+                                 (+ (frame-pixel-width exwm--floating-frame)
+                                    width-delta)
+                                 nil t))
+              (when (and (/= 0 (logand value-mask xcb:ConfigWindow:Height))
+                         (>= (abs height-delta) exwm-manage--height-delta-min))
+                (set-frame-height exwm--floating-frame
+                                  (+ (frame-pixel-height exwm--floating-frame)
+                                     height-delta)
+                                  nil t)))
+          (exwm--log "ConfigureWindow (preserve geometry)")
+          ;; Configure the unmanaged window.
+          (xcb:+request exwm--connection
+              (make-instance 'xcb:ConfigureWindow
+                             :window window
+                             :value-mask value-mask
+                             :x x :y y :width width :height height
+                             :border-width border-width
+                             :sibling sibling
+                             :stack-mode stack-mode))))))
   (xcb:flush exwm--connection))
 
 (defun exwm-manage--on-MapRequest (data _synthetic)
@@ -506,6 +562,10 @@ border-width: %d; sibling: #x%x; stack-mode: %d"
   (xcb:+event exwm--connection 'xcb:DestroyNotify
               #'exwm-manage--on-DestroyNotify))
 
+(defun exwm-manage--exit ()
+  "Exit the manage module."
+  (setq exwm-manage--_MOTIF_WM_HINTS nil))
+
 
 
 (provide 'exwm-manage)
