@@ -4,8 +4,8 @@
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 ;; Maintainer: Chris Feng <chris.w.feng@gmail.com>
-;; Version: 0.4
-;; Package-Requires: ((xelb "0.6"))
+;; Version: 0.5
+;; Package-Requires: ((xelb "0.8"))
 ;; Keywords: unix
 ;; URL: https://github.com/ch11ng/exwm
 
@@ -80,7 +80,7 @@
       (when exwm--fullscreen (exwm-layout-unset-fullscreen))
       ;; Force refresh
       (exwm-layout--refresh)
-      (exwm-input-grab-keyboard))))
+      (call-interactively #'exwm-input-grab-keyboard))))
 
 (defun exwm--update-window-type (id &optional force)
   "Update _NET_WM_WINDOW_TYPE."
@@ -326,14 +326,14 @@
                    (= action xcb:ewmh:_NET_WM_STATE_ADD))
           (dolist (f exwm-workspace--list)
             (when (equal (frame-parameter f 'exwm-outer-id) id)
-	      (exwm-layout--set-frame-fullscreen f)
-	      (xcb:+request
+              (exwm-layout--set-frame-fullscreen f)
+              (xcb:+request
                   exwm--connection
                   (make-instance 'xcb:ewmh:set-_NET_WM_STATE
                                  :window id
                                  :data (vector
                                         xcb:Atom:_NET_WM_STATE_FULLSCREEN)))
-	      (xcb:flush exwm--connection))))
+              (xcb:flush exwm--connection))))
         (when buffer                    ;ensure it's managed
           (with-current-buffer buffer
             ;; _NET_WM_STATE_MODAL
@@ -477,11 +477,15 @@
 
 (defun exwm-init (&optional frame)
   "Initialize EXWM."
-  (if (not (eq 'x (framep (or frame (selected-frame)))))
+  (if frame
+      ;; The frame might not be selected if it's created by emacslicnet.
+      (select-frame-set-input-focus frame)
+    (setq frame (selected-frame)))
+  (if (not (eq 'x (framep frame)))
       (exwm--log "Not running under X environment")
     (unless exwm--connection
       (exwm-enable 'undo)               ;never initialize again
-      (setq exwm--connection (xcb:connect-to-socket))
+      (setq exwm--connection (xcb:connect))
       (set-process-query-on-exit-flag (slot-value exwm--connection 'process)
                                       nil) ;prevent query message on exit
       (setq exwm--root
@@ -499,8 +503,8 @@
         ;; Disable some features not working well with EXWM
         (setq use-dialog-box nil)
         ;; Initialize ICCCM/EWMH support
-        ;; (xcb:icccm:init exwm--connection)
-        (xcb:ewmh:init exwm--connection)
+        (xcb:icccm:init exwm--connection t)
+        (xcb:ewmh:init exwm--connection t)
         (exwm--lock)
         (exwm--init-icccm-ewmh)
         (exwm-layout--init)
@@ -514,6 +518,26 @@
         (exwm-manage--scan)
         (run-hooks 'exwm-init-hook)))))
 
+(defvar exwm-exit-hook nil
+  "Normal hook run just before EXWM is about to exit.
+
+This hook is only run when EXWM is started with emacsclient.")
+
+(defun exwm--exit ()
+  "Exit EXWM."
+  (run-hooks 'exwm-exit-hook)
+  ;; Exit modules.
+  (exwm-input--exit)
+  (exwm-workspace--exit)
+  (exwm-manage--exit)
+  (exwm-floating--exit)
+  (exwm-layout--exit)
+  ;; Reset several import variables.
+  (setq exwm--connection nil
+        exwm--root nil
+        exwm--id-buffer-alist nil)
+  (exwm-enable))
+
 (defvar exwm-blocking-subrs '(x-file-dialog x-popup-dialog x-select-font)
   "Subrs (primitives) that would normally block EXWM.")
 
@@ -557,7 +581,7 @@
           (start-process exwm--server-name
                          nil
                          (car command-line-args) ;The executable file
-                         "-d" x-display-name
+                         "-d" (frame-parameter nil 'display)
                          "-Q"
                          (concat "--daemon=" exwm--server-name)
                          "--eval"
